#include "gameinc.h"
#include "ay8910.h"
#include "sasound.h"

static struct ROM_INFO arkanoid_roms[] =
{
  { "a75_01-1.rom", 0x8000, 0x5bcda3b0 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "a75_11.rom", 0x8000, 0xeafd7191 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "arkanoid.uc", 0x0800, 0x515d77b6 , REGION_ROM2, 0x0000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arknoidu_roms[] =
{
  { "a75-19.bin", 0x8000, 0xd3ad37d7 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "a75-18.bin", 0x8000, 0xcdc08301 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "arknoidu.uc", 0x0800, 0xde518e47, REGION_ROM2, 0x0000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkatour_roms[] =
{
  { "t_ark1.bin", 0x8000, 0xe3b8faf5 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "t_ark2.bin", 0x8000, 0x326aca4d , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "arkatour.uc", 0x0800, 0xd3249559 , REGION_ROM2, 0x0000, LOAD_NORMAL },
  { "t_ark3.bin", 0x8000, 0x5ddea3cf , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "t_ark4.bin", 0x8000, 0x5fcf2e85 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "t_ark5.bin", 0x8000, 0x7b76b192 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arknoidj_roms[] =
{
  { "a75-21.rom", 0x8000, 0xbf0455fc , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "a75-22.rom", 0x8000, 0x3a2688d3 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "arknoidj.uc", 0x0800, 0x0a4abef6 , REGION_ROM2, 0x0000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkbl2_roms[] =
{
  { "e1.6d", 0x8000, 0xdd4f2b72 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "e2.6f", 0x8000, 0xbbc33ceb , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "68705p3.6i", 0x0800, 0x389a8cfb , REGION_ROM2, 0x0000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkbl3_roms[] =
{
  { "arkanunk.1", 0x8000, 0xb0f73900 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "arkanunk.2", 0x8000, 0x9827f297 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO paddle2_roms[] =
{
  { "paddle2.16", 0x8000, 0xa286333c , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "paddle2.17", 0x8000, 0x04c2acb5 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkatayt_roms[] =
{
  { "arkanoid.1", 0x8000, 0x6e0a2b6f , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "arkanoid.2", 0x8000, 0x5a97dd56 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkblock_roms[] =
{
  { "block01.bin", 0x8000, 0x5be667e1 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "block02.bin", 0x8000, 0x4f883ef1 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkbloc2_roms[] =
{
  { "ark-6.bin", 0x8000, 0x0be015de , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "arkgc.2", 0x8000, 0x9f0d4754 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct ROM_INFO arkangc_roms[] =
{
  { "arkgc.1", 0x8000, 0xc54232e6 , REGION_ROM1, 0x0000, LOAD_NORMAL },
  { "arkgc.2", 0x8000, 0x9f0d4754 , REGION_ROM1, 0x8000, LOAD_NORMAL },
  { "a75_03.rom", 0x8000, 0x038b74ba , REGION_GFX1, 0x00000, LOAD_NORMAL },
  { "a75_04.rom", 0x8000, 0x71fae199 , REGION_GFX1, 0x08000, LOAD_NORMAL },
  { "a75_05.rom", 0x8000, 0xc76374e2 , REGION_GFX1, 0x10000, LOAD_NORMAL },
  { "07.bpr", 0x0200, 0x0af8b289 , REGION_PROMS, 0x0000, LOAD_NORMAL },
  { "08.bpr", 0x0200, 0xabb002fb , REGION_PROMS, 0x0200, LOAD_NORMAL },
  { "09.bpr", 0x0200, 0xa7c6c277 , REGION_PROMS, 0x0400, LOAD_NORMAL },
  { NULL, 0, 0, 0, 0, 0 }
};

static struct DSW_DATA dsw_data_arkanoid_4[] =
{
  { "Allow Continue", 0x01, 2 },
  { MSG_NO, 0x01, 0x00 },
  { MSG_YES, 0x00, 0x00 },
  { MSG_SCREEN, 0x02, 2 },
  { MSG_OFF, 0x02, 0x00 },
  { MSG_ON, 0x00, 0x00 },
  { MSG_SERVICE, 0x04,2 },
  { MSG_ON, 0,0 },
  { MSG_OFF, 0x04,0 },
  { MSG_DIFFICULTY, 0x08, 2 },
  { "Easy" , 0x08, 0x00 },
  { "Hard" , 0x00, 0x00 },
  { MSG_EXTRA_LIFE, 0x10, 2 },
  { "20K60K and every 60K" , 0x10, 0x00 },
  { "20K only" , 0x00, 0x00 },
  { MSG_LIVES, 0x20, 2 },
  { "3" , 0x20, 0x00 },
  { "5" , 0x00, 0x00 },
  { MSG_COINAGE, 0xc0, 4 },
  { MSG_2COIN_1PLAY, 0x40, 0x00 },
  { MSG_1COIN_1PLAY, 0xc0, 0x00 },
  { MSG_1COIN_1PLAY, 0x80, 0x00 },
  { MSG_1COIN_6PLAY, 0x00, 0x00 },
  { NULL, 0, 0}
};

static struct DSW_INFO arkanoid_dsw[] =
{
  { 8, 0xfe, dsw_data_arkanoid_4 },
  { 0, 0, NULL }
};

static struct DSW_DATA dsw_data_arknoidj_4[] =
{
  { "Allow Continue", 0x01, 2 },
  { MSG_NO, 0x01, 0x00 },
  { MSG_YES, 0x00, 0x00 },
  { MSG_SCREEN, 0x02, 2 },
  { MSG_OFF, 0x02, 0x00 },
  { MSG_ON, 0x00, 0x00 },
  { MSG_SERVICE, 0x04,2 },
  { MSG_ON, 0,0 },
  { MSG_OFF, 0x04,0 },
  { MSG_DIFFICULTY, 0x08, 2 },
  { "Easy" , 0x08, 0x00 },
  { "Hard" , 0x00, 0x00 },
  { MSG_EXTRA_LIFE, 0x10, 2 },
  { "20K60K and every 60K" , 0x10, 0x00 },
  { "20K only" , 0x00, 0x00 },
  { MSG_LIVES, 0x20, 2 },
  { "3" , 0x20, 0x00 },
  { "5" , 0x00, 0x00 },
  { MSG_COINAGE, 0x40, 2 },
  { MSG_1COIN_1PLAY, 0x40, 0x00 },
  { MSG_1COIN_1PLAY, 0x00, 0x00 },
  { MSG_CABINET, 0x80, 2 },
  { MSG_UPRIGHT, 0x00, 0x00 },
  { MSG_TABLE, 0x80, 0x00 },
  { NULL, 0, 0}
};

static struct DSW_INFO arknoidj_dsw[] =
{
  { 8, 0x7e, dsw_data_arknoidj_4 },
  { 0, 0, NULL }
};

static struct DSW_DATA dsw_data_arkatayt_4[] =
{
  { "Allow Continue", 0x01, 2 },
  { MSG_NO, 0x01, 0x00 },
  { MSG_YES, 0x00, 0x00 },
  { MSG_SCREEN, 0x02, 2 },
  { MSG_OFF, 0x02, 0x00 },
  { MSG_ON, 0x00, 0x00 },
  { MSG_SERVICE, 0x04,2 },
  { MSG_ON, 0,0 },
  { MSG_OFF, 0x04,0 },
  { MSG_DIFFICULTY, 0x08, 2 },
  { "Easy" , 0x08, 0x00 },
  { "Hard" , 0x00, 0x00 },
  { MSG_EXTRA_LIFE, 0x10, 2 },
  { "60K100K and every 60K" , 0x10, 0x00 },
  { "60K only" , 0x00, 0x00 },
  { MSG_LIVES, 0x20, 2 },
  { "2" , 0x20, 0x00 },
  { "3" , 0x00, 0x00 },
  { MSG_COINAGE, 0x40, 2 },
  { MSG_2COIN_1PLAY, 0x00, 0x00 },
  { MSG_1COIN_1PLAY, 0x40, 0x00 },
  { MSG_CABINET, 0x80, 2 },
  { MSG_UPRIGHT, 0x00, 0x00 },
  { MSG_TABLE, 0x80, 0x00 },
  { NULL, 0, 0}
};

static struct DSW_INFO arkatayt_dsw[] =
{
  { 8, 0x5e, dsw_data_arkatayt_4 },
  { 0, 0, NULL }
};

static struct INPUT_INFO arkanoid_inputs[] =
{
  { KB_DEF_P1_START, MSG_P1_START, 0x0000, 0x01, BIT_ACTIVE_0 },
  { KB_DEF_P2_START, MSG_P2_START, 0x0000, 0x02, BIT_ACTIVE_0 },
  { KB_DEF_SERVICE,  MSG_SERVICE,  0x0000, 0x04, BIT_ACTIVE_0 },
  { KB_DEF_TILT,     MSG_TILT,     0x0000, 0x08, BIT_ACTIVE_0 },
  { KB_DEF_COIN1,    MSG_COIN1,    0x0000, 0x10, BIT_ACTIVE_1 },
  { KB_DEF_COIN2,    MSG_COIN2,    0x0000, 0x20, BIT_ACTIVE_1 },
  // The last 2 bits are from the 68705 (mcu)
  { KB_DEF_COIN3,    MSG_COIN3,    0x0000, 0x40, BIT_ACTIVE_1 }, // need to be 1 for some bootlegs
  { KB_DEF_COIN3,    MSG_COIN3,    0x0000, 0x80, BIT_ACTIVE_1 }, // need to be 1 for some bootlegs

  { KB_DEF_P1_B1, MSG_P1_B1, 0x0002, 0x01, BIT_ACTIVE_0 },
  { KB_DEF_P2_B1, MSG_P2_B1, 0x0002, 0x04, BIT_ACTIVE_0 },
};

static struct INPUT_INFO arknoidj_inputs[] =
{
  { KB_DEF_P1_START, MSG_P1_START, 0x00, 0x01, BIT_ACTIVE_0 },
  { KB_DEF_P2_START, MSG_P2_START, 0x00, 0x02, BIT_ACTIVE_0 },
  { KB_DEF_SERVICE, MSG_SERVICE, 0x00, 0x04, BIT_ACTIVE_0 },
  { KB_DEF_TILT, MSG_TILT, 0x00, 0x08, BIT_ACTIVE_0 },
  { KB_DEF_COIN1, MSG_COIN1, 0x00, 0x10, BIT_ACTIVE_1 },
  { KB_DEF_COIN2, MSG_COIN2, 0x00, 0x20, BIT_ACTIVE_1 },
  { KB_DEF_COIN3,    MSG_COIN3,    0x0000, 0x40, BIT_ACTIVE_0 }, // need to be 1 for some bootlegs
  { KB_DEF_COIN3,    MSG_COIN3,    0x0000, 0x80, BIT_ACTIVE_1 }, // need to be 1 for some bootlegs

  { KB_DEF_P1_B1, MSG_P1_B1, 0x02, 0x01, BIT_ACTIVE_0 },
  { KB_DEF_P2_B1, MSG_P2_B1, 0x02, 0x04, BIT_ACTIVE_0 },
};

static struct GFX_LAYOUT charlayout =
{
	8,8,	/* 8*8 characters */
	4096,	/* 4096 characters */
	3,	/* 3 bits per pixel */
	{ 2*4096*8*8, 4096*8*8, 0 },	/* the two bitplanes are separated */
	{ 0, 1, 2, 3, 4, 5, 6, 7 },
	{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
	8*8	/* every char takes 8 consecutive bytes */
};

static struct GFX_LIST arkanoid_gfx[] =
{
	{ REGION_GFX1, &charlayout, },
	/* sprites use the same characters above, but are 16x8 */
	{ -1 } /* end of array */
};

UINT8 input_port_4_r(UINT16 offset) {
  return get_dsw(0);
}

static struct AY8910interface ay8910_interface =
{
	1,	/* 1 chips */
	1500000,	/* 1.5 MHz ???? */
	{ 33 },
	{ 0 },
	{ input_port_4_r },
	{ 0 },
	{ 0 }
};

static struct SOUND_INFO arkanoid_sound[] =
{
   { SOUND_AY8910,  &ay8910_interface,  },
   { 0,             NULL,               },
};

static void load_arkanoid() {

  if (!(RAM = AllocateMem(0x2800))) return;
  memset(RAM,0,0x2800);

  GFX_BG0_SOLID = make_solid_region_8x8(REGION_GFX1);
  GFX_OBJ_SOLID = make_solid_region_16x8(REGION_GFX1);

  AddZ80AROMBase(Z80ROM, 0x0038, 0x0066); // ???!
  AddZ80AReadByte(0x0000,0xbfff, NULL, ROM);
  AddZ80AReadByte(0xc000,0xc7ff, NULL, RAM);
  AddZ80AReadByte(0xd001,0xd001, AY8910_read_port_0_r, NULL);
  AddZ80AReadByte(0xd00c,0xd018, input_stuff, NULL);
  AddZ80AReadByte(0xe000,0xffff, NULL, RAM+0x800);

  AddZ80AWriteByte(0xc000, 0xc7ff, NULL, RAM);
  AddZ80AWriteByte(0xe000, 0xffff, NULL, RAM+0x800);
  AddZ80AWriteByte(0xd000, 0xd000, AY8910_control_port_0_w, NULL);
  AddZ80AWriteByte(0xd001, 0xd001, AY8910_write_port_0_w,   NULL);

  AddZ80AWriteByte(0xd008, 0xd018, output_stuff,   NULL);

}

static void execute_arkanoid() {
  cpu_execute_cycles(CPU_Z80_0, CPU_FRAME_MHz(6,60) );
  cpu_interrupt(audio_cpu, 0x38);
}

static struct DIR_INFO arkanoid_dirs[] =
{
   { "arkanoid", },
   { NULL, },
};

static void draw_arkanoid() {

}

static struct VIDEO_INFO arkanoid_video =
{
   draw_arkanoid,
   256,
   224,
   32,
   VIDEO_ROTATE_270 |
   VIDEO_ROTATABLE,
   arkanoid_gfx,
};

GAME( arkatayt ,
   arkatayt_dirs,
   arkatayt_roms,
   arknoidj_inputs,
   arkatayt_dsw,
   NULL,

   load_arkanoid,
   NULL,
   &arkanoid_video,
   execute_arkanoid,
   "arkanoid",
   "Arkanoid (Tayto bootleg, Japanese)",
   "Arkanoid",
   COMPANY_ID_TAITO,
   NULL,
   1986,
   arkanoid_sound,
   GAME_BREAKOUT
);
